"""Constants."""
from __future__ import annotations

__all__ = ('BLURAY_DUAL_LAYER_SIZE_BYTES', 'BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_QUADRUPLE_LAYER_SIZE_BYTES', 'BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_SINGLE_LAYER_SIZE_BYTES', 'BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED',
           'BLURAY_TRIPLE_LAYER_SIZE_BYTES', 'BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED',
           'CD_R_BYTES', 'CD_R_BYTES_ADJUSTED', 'DVD_R_DUAL_LAYER_SIZE_BYTES',
           'DVD_R_SINGLE_LAYER_SIZE_BYTES')

CD_R_BYTES = 737_280_000
"""700 MB CD-R size in bytes."""
CD_R_BYTES_ADJUSTED = CD_R_BYTES - 5_242_880
"""700 MB CD-R size in bytes with 5 MiB free space."""
DVD_R_SINGLE_LAYER_SIZE_BYTES = 4_707_319_808
"""DVD-R size in bytes."""
DVD_R_DUAL_LAYER_SIZE_BYTES = 8_543_666_176
"""DVD-R DL size in bytes."""
FREE_SPACE_BYTES = 52_428_800
"""50 MiB free space."""
DVD_R_SINGLE_LAYER_SIZE_BYTES_ADJUSTED = DVD_R_SINGLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""DVD-R size in bytes with 50 MiB left empty."""
DVD_R_DUAL_LAYER_SIZE_BYTES_ADJUSTED = DVD_R_DUAL_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""DVD-R DL size in bytes with 50 MiB left empty."""
BLURAY_SINGLE_LAYER_SIZE_BYTES = 25_025_314_816
"""25 GB Blu-Ray disc size."""
BLURAY_DUAL_LAYER_SIZE_BYTES = 50_050_629_632
"""50 GB Blu-Ray disc size."""
BLURAY_TRIPLE_LAYER_SIZE_BYTES = 100_103_356_416
"""100 GB Blu-Ray disc size."""
BLURAY_QUADRUPLE_LAYER_SIZE_BYTES = 128_001_769_472
"""128 GB Blu-Ray disc size in bytes."""
BLURAY_SINGLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_SINGLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""25 GB disc size in bytes with 50 MiB left empty."""
BLURAY_DUAL_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_DUAL_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""50 GB disc size in bytes with 50 MiB left empty."""
BLURAY_TRIPLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_TRIPLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""100 GB disc size in bytes with 50 MiB left empty."""
BLURAY_QUADRUPLE_LAYER_SIZE_BYTES_ADJUSTED = BLURAY_QUADRUPLE_LAYER_SIZE_BYTES - FREE_SPACE_BYTES
"""128 GB disc size in bytes with 50 MiB left empty."""
